@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2021, MariaDB Corporation.
+Copyright (c) 2017, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -30,7 +30,9 @@ Created 11/26/1995 Heikki Tuuri
 #include "buf0flu.h"
 #include "page0types.h"
 #include "mtr0log.h"
+#include "log0log.h"
 #include "row0trunc.h"
+
 #include "log0recv.h"
 
 /** Iterate over a memo block in reverse. */
@@ -202,111 +204,154 @@ private:
 
 /** Release latches and decrement the buffer fix count.
 @param slot	memo slot */
-static void memo_slot_release(mtr_memo_slot_t *slot)
+static
+void
+memo_slot_release(mtr_memo_slot_t* slot)
 {
-  switch (slot->type) {
+	switch (slot->type) {
+	case MTR_MEMO_BUF_FIX:
+	case MTR_MEMO_PAGE_S_FIX:
+	case MTR_MEMO_PAGE_SX_FIX:
+	case MTR_MEMO_PAGE_X_FIX: {
+
+		buf_block_t*	block;
+
+		block = reinterpret_cast<buf_block_t*>(slot->object);
+
+		buf_block_unfix(block);
+		buf_page_release_latch(block, slot->type);
+		break;
+	}
+
+	case MTR_MEMO_S_LOCK:
+		rw_lock_s_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
+		break;
+
+	case MTR_MEMO_SX_LOCK:
+		rw_lock_sx_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
+		break;
+
+	case MTR_MEMO_X_LOCK:
+		rw_lock_x_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
+		break;
+
 #ifdef UNIV_DEBUG
-  default:
-    ut_ad(!"invalid type");
-    break;
-  case MTR_MEMO_MODIFY:
-    break;
+	default:
+		ut_ad(slot->type == MTR_MEMO_MODIFY);
 #endif /* UNIV_DEBUG */
-  case MTR_MEMO_S_LOCK:
-    rw_lock_s_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
-    break;
-  case MTR_MEMO_SX_LOCK:
-    rw_lock_sx_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
-    break;
-  case MTR_MEMO_SPACE_X_LOCK:
-    {
-      fil_space_t *space= static_cast<fil_space_t*>(slot->object);
-      space->committed_size= space->size;
-      rw_lock_x_unlock(&space->latch);
-    }
-    break;
-  case MTR_MEMO_X_LOCK:
-    rw_lock_x_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
-    break;
-  case MTR_MEMO_BUF_FIX:
-  case MTR_MEMO_PAGE_S_FIX:
-  case MTR_MEMO_PAGE_SX_FIX:
-  case MTR_MEMO_PAGE_X_FIX:
-    buf_block_t *block= reinterpret_cast<buf_block_t*>(slot->object);
-    buf_page_release_latch(block, slot->type);
-    buf_block_unfix(block);
-    break;
-  }
-  slot->object= NULL;
+	}
+
+	slot->object = NULL;
+}
+
+/** Unfix a page, do not release the latches on the page.
+@param slot	memo slot */
+static
+void
+memo_block_unfix(mtr_memo_slot_t* slot)
+{
+	switch (slot->type) {
+	case MTR_MEMO_BUF_FIX:
+	case MTR_MEMO_PAGE_S_FIX:
+	case MTR_MEMO_PAGE_X_FIX:
+	case MTR_MEMO_PAGE_SX_FIX: {
+		buf_block_unfix(reinterpret_cast<buf_block_t*>(slot->object));
+		break;
+	}
+
+	case MTR_MEMO_S_LOCK:
+	case MTR_MEMO_X_LOCK:
+	case MTR_MEMO_SX_LOCK:
+		break;
+#ifdef UNIV_DEBUG
+	default:
+#endif /* UNIV_DEBUG */
+		break;
+	}
+}
+/** Release latches represented by a slot.
+@param slot	memo slot */
+static
+void
+memo_latch_release(mtr_memo_slot_t* slot)
+{
+	switch (slot->type) {
+	case MTR_MEMO_BUF_FIX:
+	case MTR_MEMO_PAGE_S_FIX:
+	case MTR_MEMO_PAGE_SX_FIX:
+	case MTR_MEMO_PAGE_X_FIX: {
+		buf_block_t*	block;
+
+		block = reinterpret_cast<buf_block_t*>(slot->object);
+
+		memo_block_unfix(slot);
+
+		buf_page_release_latch(block, slot->type);
+
+		slot->object = NULL;
+		break;
+	}
+
+	case MTR_MEMO_S_LOCK:
+		rw_lock_s_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
+		slot->object = NULL;
+		break;
+
+	case MTR_MEMO_X_LOCK:
+		rw_lock_x_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
+		slot->object = NULL;
+		break;
+
+	case MTR_MEMO_SX_LOCK:
+		rw_lock_sx_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
+		slot->object = NULL;
+		break;
+
+#ifdef UNIV_DEBUG
+	default:
+		ut_ad(slot->type == MTR_MEMO_MODIFY);
+
+		slot->object = NULL;
+#endif /* UNIV_DEBUG */
+	}
 }
 
 /** Release the latches acquired by the mini-transaction. */
 struct ReleaseLatches {
-  /** @return true always. */
-  bool operator()(mtr_memo_slot_t *slot) const
-  {
-    if (!slot->object)
-      return true;
-    switch (slot->type) {
-#ifdef UNIV_DEBUG
-    default:
-      ut_ad(!"invalid type");
-      break;
-    case MTR_MEMO_MODIFY:
-      break;
-#endif /* UNIV_DEBUG */
-    case MTR_MEMO_S_LOCK:
-      rw_lock_s_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
-      break;
-    case MTR_MEMO_SPACE_X_LOCK:
-      {
-        fil_space_t *space= static_cast<fil_space_t*>(slot->object);
-        space->committed_size= space->size;
-        rw_lock_x_unlock(&space->latch);
-      }
-      break;
-    case MTR_MEMO_X_LOCK:
-      rw_lock_x_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
-      break;
-    case MTR_MEMO_SX_LOCK:
-      rw_lock_sx_unlock(reinterpret_cast<rw_lock_t*>(slot->object));
-      break;
-    case MTR_MEMO_BUF_FIX:
-    case MTR_MEMO_PAGE_S_FIX:
-    case MTR_MEMO_PAGE_SX_FIX:
-    case MTR_MEMO_PAGE_X_FIX:
-      buf_block_t *block= reinterpret_cast<buf_block_t*>(slot->object);
-      buf_page_release_latch(block, slot->type);
-      buf_block_unfix(block);
-      break;
-    }
-    slot->object= NULL;
-    return true;
-  }
+
+	/** @return true always. */
+	bool operator()(mtr_memo_slot_t* slot) const
+	{
+		if (slot->object != NULL) {
+			memo_latch_release(slot);
+		}
+
+		return(true);
+	}
 };
 
 /** Release the latches and blocks acquired by the mini-transaction. */
 struct ReleaseAll {
-  /** @return true always. */
-  bool operator()(mtr_memo_slot_t *slot) const
-  {
-    if (slot->object)
-      memo_slot_release(slot);
-    return true;
-  }
+	/** @return true always. */
+	bool operator()(mtr_memo_slot_t* slot) const
+	{
+		if (slot->object != NULL) {
+			memo_slot_release(slot);
+		}
+
+		return(true);
+	}
 };
 
-#ifdef UNIV_DEBUG
 /** Check that all slots have been handled. */
 struct DebugCheck {
 	/** @return true always. */
 	bool operator()(const mtr_memo_slot_t* slot) const
 	{
-		ut_ad(!slot->object);
+		ut_a(slot->object == NULL);
 		return(true);
 	}
 };
-#endif
 
 /** Release a resource acquired by the mini-transaction. */
 struct ReleaseBlocks {
@@ -359,6 +404,59 @@ struct ReleaseBlocks {
 	FlushObserver*	m_flush_observer;
 };
 
+class mtr_t::Command {
+public:
+	/** Constructor.
+	Takes ownership of the mtr->m_impl, is responsible for deleting it.
+	@param[in,out]	mtr	mini-transaction */
+	explicit Command(mtr_t* mtr) : m_impl(&mtr->m_impl), m_locks_released()
+	{}
+
+	/** Destructor */
+	~Command()
+	{
+		ut_ad(m_impl == 0);
+	}
+
+	/** Write the redo log record, add dirty pages to the flush list and
+	release the resources. */
+	void execute();
+
+	/** Release the blocks used in this mini-transaction. */
+	void release_blocks();
+
+	/** Release the latches acquired by the mini-transaction. */
+	void release_latches();
+
+	/** Release both the latches and blocks used in the mini-transaction. */
+	void release_all();
+
+	/** Release the resources */
+	void release_resources();
+
+	/** Append the redo log records to the redo log buffer.
+	@param[in]	len	number of bytes to write */
+	void finish_write(ulint len);
+
+private:
+	/** Prepare to write the mini-transaction log to the redo log buffer.
+	@return number of bytes to write in finish_write() */
+	ulint prepare_write();
+
+	/** The mini-transaction state. */
+	mtr_t::Impl*		m_impl;
+
+	/** Set to 1 after the user thread releases the latches. The log
+	writer thread must wait for this to be set to 1. */
+	volatile ulint		m_locks_released;
+
+	/** Start lsn of the possible log entry for this mtr */
+	lsn_t			m_start_lsn;
+
+	/** End lsn of the possible log entry for this mtr */
+	lsn_t			m_end_lsn;
+};
+
 /** Write the block contents to the REDO log */
 struct mtr_write_log_t {
 	/** Append a block to the redo log buffer.
@@ -382,7 +480,7 @@ mtr_write_log(
 	ut_ad(!recv_no_log_write);
 	DBUG_PRINT("ib_log",
 		   (ULINTPF " extra bytes written at " LSN_PF,
-		    len, log_sys->lsn));
+		    len, log_sys.lsn));
 
 	log_reserve_and_open(len);
 	log->for_each_block(write_log);
@@ -392,161 +490,76 @@ mtr_write_log(
 /** Start a mini-transaction. */
 void mtr_t::start()
 {
-  MEM_UNDEFINED(this, sizeof *this);
+	UNIV_MEM_INVALID(this, sizeof(*this));
 
-  new(&m_memo) mtr_buf_t();
-  new(&m_log) mtr_buf_t();
+	UNIV_MEM_INVALID(&m_impl, sizeof(m_impl));
 
-  m_made_dirty= false;
-  m_inside_ibuf= false;
-  m_modifications= false;
-  m_n_log_recs= 0;
-  m_log_mode= MTR_LOG_ALL;
-  ut_d(m_user_space_id= TRX_SYS_SPACE);
-  m_user_space= NULL;
-  m_undo_space= NULL;
-  m_sys_space= NULL;
-  m_state= MTR_STATE_ACTIVE;
-  m_flush_observer= NULL;
-  m_commit_lsn= 0;
+	m_commit_lsn = 0;
+
+	new(&m_impl.m_log) mtr_buf_t();
+	new(&m_impl.m_memo) mtr_buf_t();
+
+	m_impl.m_mtr = this;
+	m_impl.m_log_mode = MTR_LOG_ALL;
+	m_impl.m_inside_ibuf = false;
+	m_impl.m_modifications = false;
+	m_impl.m_made_dirty = false;
+	m_impl.m_n_log_recs = 0;
+	m_impl.m_state = MTR_STATE_ACTIVE;
+	ut_d(m_impl.m_user_space_id = TRX_SYS_SPACE);
+	m_impl.m_user_space = NULL;
+	m_impl.m_flush_observer = NULL;
+
+	ut_d(m_impl.m_magic_n = MTR_MAGIC_N);
 }
 
 /** Release the resources */
-inline void mtr_t::release_resources()
+void
+mtr_t::Command::release_resources()
 {
-  ut_d(m_memo.for_each_block_in_reverse(CIterate<DebugCheck>()));
-  m_log.erase();
-  m_memo.erase();
-  m_state= MTR_STATE_COMMITTED;
+	ut_ad(m_impl->m_magic_n == MTR_MAGIC_N);
+
+	/* Currently only used in commit */
+	ut_ad(m_impl->m_state == MTR_STATE_COMMITTING);
+
+	ut_d(m_impl->m_memo.for_each_block_in_reverse(CIterate<DebugCheck>()));
+
+	/* Reset the mtr buffers */
+	m_impl->m_log.erase();
+
+	m_impl->m_memo.erase();
+
+	m_impl->m_state = MTR_STATE_COMMITTED;
+
+	m_impl = 0;
 }
 
 /** Commit a mini-transaction. */
 void
 mtr_t::commit()
 {
-  ut_ad(is_active());
-  ut_ad(!is_inside_ibuf());
+	ut_ad(is_active());
+	ut_ad(!is_inside_ibuf());
+	ut_ad(m_impl.m_magic_n == MTR_MAGIC_N);
+	m_impl.m_state = MTR_STATE_COMMITTING;
 
-  /* This is a dirty read, for debugging. */
-  ut_ad(!m_modifications || !recv_no_log_write);
-  ut_ad(!m_modifications || m_log_mode != MTR_LOG_NONE);
+	/* This is a dirty read, for debugging. */
+	ut_ad(!m_impl.m_modifications || !recv_no_log_write);
 
-  if (m_modifications
-      && (m_n_log_recs || m_log_mode == MTR_LOG_NO_REDO))
-  {
-    ut_ad(!srv_read_only_mode || m_log_mode == MTR_LOG_NO_REDO);
+	Command	cmd(this);
 
-    lsn_t start_lsn;
+	if (m_impl.m_modifications
+	    && (m_impl.m_n_log_recs > 0
+		|| m_impl.m_log_mode == MTR_LOG_NO_REDO)) {
 
-    if (const ulint len= prepare_write())
-      start_lsn= finish_write(len);
-    else
-      start_lsn= m_commit_lsn;
+		ut_ad(!srv_read_only_mode
+		      || m_impl.m_log_mode == MTR_LOG_NO_REDO);
 
-    if (m_made_dirty)
-      log_flush_order_mutex_enter();
-
-    /* It is now safe to release the log mutex because the
-    flush_order mutex will ensure that we are the first one
-    to insert into the flush list. */
-    log_mutex_exit();
-
-    m_memo.for_each_block_in_reverse(CIterate<const ReleaseBlocks>
-                                     (ReleaseBlocks(start_lsn, m_commit_lsn,
-                                                    m_flush_observer)));
-    if (m_made_dirty)
-      log_flush_order_mutex_exit();
-
-    m_memo.for_each_block_in_reverse(CIterate<ReleaseLatches>());
-  }
-  else
-    m_memo.for_each_block_in_reverse(CIterate<ReleaseAll>());
-
-  release_resources();
-}
-
-#ifdef UNIV_DEBUG
-/** Check that all pages belong to a shrunk tablespace. */
-struct Shrink
-{
-  const fil_space_t &space;
-  Shrink(const fil_space_t &space) : space(space) {}
-
-  bool operator()(const mtr_memo_slot_t *slot) const
-  {
-    if (!slot->object)
-      return true;
-    switch (slot->type) {
-    default:
-      ut_ad("invalid type" == 0);
-      return false;
-    case MTR_MEMO_MODIFY:
-      break;
-    case MTR_MEMO_SPACE_X_LOCK:
-      ut_ad(&space == slot->object);
-      return true;
-    case MTR_MEMO_PAGE_X_FIX:
-    case MTR_MEMO_PAGE_SX_FIX:
-      const buf_page_t &bpage= static_cast<buf_block_t*>(slot->object)->page;
-      const page_id_t &id= bpage.id;
-      if (id.space() == 0 && id.page_no() == TRX_SYS_PAGE_NO)
-      {
-        ut_ad(srv_is_undo_tablespace(space.id));
-        break;
-      }
-      ut_ad(id.space() == space.id);
-      ut_ad(id.page_no() < space.size);
-      ut_ad(bpage.state == BUF_BLOCK_FILE_PAGE);
-      ut_ad(!bpage.oldest_modification);
-      break;
-    }
-    return true;
-  }
-};
-#endif
-
-/** Commit a mini-transaction that is shrinking a tablespace.
-@param space   tablespace that is being shrunk */
-void mtr_t::commit_shrink(fil_space_t &space)
-{
-  ut_ad(is_active());
-  ut_ad(!is_inside_ibuf());
-  ut_ad(!high_level_read_only);
-  ut_ad(m_modifications);
-  ut_ad(m_made_dirty);
-  ut_ad(!recv_recovery_is_on());
-  ut_ad(m_log_mode == MTR_LOG_ALL);
-  ut_ad(UT_LIST_GET_LEN(space.chain) == 1);
-
-  log_write_and_flush_prepare();
-
-  const lsn_t start_lsn= finish_write(prepare_write());
-
-  log_flush_order_mutex_enter();
-  /* Durably write the reduced FSP_SIZE before truncating the data file. */
-  log_write_and_flush();
-
-  os_file_truncate(space.chain.start->name, space.chain.start->handle,
-                   os_offset_t(space.size) << srv_page_size_shift, true);
-
-  ut_d(m_memo.for_each_block_in_reverse(CIterate<Shrink>(space)));
-
-  m_memo.for_each_block_in_reverse(CIterate<const ReleaseBlocks>
-                                   (ReleaseBlocks(start_lsn, m_commit_lsn,
-                                                  m_flush_observer)));
-  log_flush_order_mutex_exit();
-
-  mutex_enter(&fil_system->mutex);
-  ut_ad(space.is_being_truncated);
-  ut_ad(space.stop_new_ops);
-  space.stop_new_ops= false;
-  space.is_being_truncated= false;
-  mutex_exit(&fil_system->mutex);
-
-  m_memo.for_each_block_in_reverse(CIterate<ReleaseLatches>());
-  srv_stats.log_write_requests.inc();
-
-  release_resources();
+		cmd.execute();
+	} else {
+		cmd.release_all();
+		cmd.release_resources();
+	}
 }
 
 /** Commit a mini-transaction that did not modify any pages,
@@ -565,36 +578,40 @@ mtr_t::commit_checkpoint(
 	ut_ad(log_mutex_own());
 	ut_ad(is_active());
 	ut_ad(!is_inside_ibuf());
+	ut_ad(m_impl.m_magic_n == MTR_MAGIC_N);
 	ut_ad(get_log_mode() == MTR_LOG_ALL);
-	ut_ad(!m_made_dirty);
-	ut_ad(m_memo.size() == 0);
+	ut_ad(!m_impl.m_made_dirty);
+	ut_ad(m_impl.m_memo.size() == 0);
 	ut_ad(!srv_read_only_mode);
-	ut_ad(write_mlog_checkpoint || m_n_log_recs > 1);
+	ut_d(m_impl.m_state = MTR_STATE_COMMITTING);
+	ut_ad(write_mlog_checkpoint || m_impl.m_n_log_recs > 1);
 
-	switch (m_n_log_recs) {
+	switch (m_impl.m_n_log_recs) {
 	case 0:
 		break;
 	case 1:
-		*m_log.front()->begin() |= MLOG_SINGLE_REC_FLAG;
+		*m_impl.m_log.front()->begin() |= MLOG_SINGLE_REC_FLAG;
 		break;
 	default:
-		mlog_catenate_ulint(&m_log, MLOG_MULTI_REC_END, MLOG_1BYTE);
+		mlog_catenate_ulint(
+			&m_impl.m_log, MLOG_MULTI_REC_END, MLOG_1BYTE);
 	}
 
 	if (write_mlog_checkpoint) {
-		byte*	ptr = m_log.push<byte*>(SIZE_OF_MLOG_CHECKPOINT);
+		byte*	ptr = m_impl.m_log.push<byte*>(SIZE_OF_MLOG_CHECKPOINT);
 		compile_time_assert(SIZE_OF_MLOG_CHECKPOINT == 1 + 8);
 		*ptr = MLOG_CHECKPOINT;
 		mach_write_to_8(ptr + 1, checkpoint_lsn);
 	}
 
-	finish_write(m_log.size());
-	release_resources();
+	Command	cmd(this);
+	cmd.finish_write(m_impl.m_log.size());
+	cmd.release_resources();
 
 	if (write_mlog_checkpoint) {
 		DBUG_PRINT("ib_log",
 			   ("MLOG_CHECKPOINT(" LSN_PF ") written at " LSN_PF,
-			    checkpoint_lsn, log_sys->lsn));
+			    checkpoint_lsn, log_sys.lsn));
 	}
 }
 
@@ -606,12 +623,8 @@ mtr_t::commit_checkpoint(
 bool
 mtr_t::is_named_space(ulint space) const
 {
-	ut_ad(!m_sys_space || m_sys_space->id == TRX_SYS_SPACE);
-	ut_ad(!m_undo_space || m_undo_space->id != TRX_SYS_SPACE);
-	ut_ad(!m_user_space || m_user_space->id != TRX_SYS_SPACE);
-	ut_ad(!m_sys_space || m_sys_space != m_user_space);
-	ut_ad(!m_sys_space || m_sys_space != m_undo_space);
-	ut_ad(!m_user_space || m_user_space != m_undo_space);
+	ut_ad(!m_impl.m_user_space
+	      || m_impl.m_user_space->id != TRX_SYS_SPACE);
 
 	switch (get_log_mode()) {
 	case MTR_LOG_NONE:
@@ -619,13 +632,35 @@ mtr_t::is_named_space(ulint space) const
 		return(true);
 	case MTR_LOG_ALL:
 	case MTR_LOG_SHORT_INSERTS:
-		return(m_user_space_id == space
+		return(m_impl.m_user_space_id == space
 		       || is_predefined_tablespace(space));
 	}
 
 	ut_error;
 	return(false);
 }
+/** Check if a tablespace is associated with the mini-transaction
+(needed for generating a MLOG_FILE_NAME record)
+@param[in]	space	tablespace
+@return whether the mini-transaction is associated with the space */
+bool mtr_t::is_named_space(const fil_space_t* space) const
+{
+	ut_ad(!m_impl.m_user_space
+	      || m_impl.m_user_space->id != TRX_SYS_SPACE);
+
+	switch (get_log_mode()) {
+	case MTR_LOG_NONE:
+	case MTR_LOG_NO_REDO:
+		return true;
+	case MTR_LOG_ALL:
+	case MTR_LOG_SHORT_INSERTS:
+		return(m_impl.m_user_space == space
+		       || is_predefined_tablespace(space->id));
+	}
+
+	ut_error;
+	return false;
+}
 #endif /* UNIV_DEBUG */
 
 /** Acquire a tablespace X-latch.
@@ -639,97 +674,45 @@ mtr_t::x_lock_space(ulint space_id, const char* file, unsigned line)
 {
 	fil_space_t*	space;
 
+	ut_ad(m_impl.m_magic_n == MTR_MAGIC_N);
 	ut_ad(is_active());
 
 	if (space_id == TRX_SYS_SPACE) {
-		space = m_sys_space;
-
-		if (!space) {
-			space = m_sys_space = fil_space_get(space_id);
-		}
-	} else if ((space = m_user_space) && space_id == space->id) {
-	} else if ((space = m_undo_space) && space_id == space->id) {
-	} else if (get_log_mode() == MTR_LOG_NO_REDO) {
-		space = fil_space_get(space_id);
-		ut_ad(space->purpose == FIL_TYPE_TEMPORARY
-		      || space->purpose == FIL_TYPE_IMPORT
-		      || space->redo_skipped_count > 0
-		      || srv_is_tablespace_truncated(space->id));
+		space = fil_system.sys_space;
+	} else if ((space = m_impl.m_user_space) && space_id == space->id) {
 	} else {
-		/* called from trx_rseg_create() */
-		space = m_undo_space = fil_space_get(space_id);
+		space = fil_space_get(space_id);
+		ut_ad(get_log_mode() != MTR_LOG_NO_REDO
+		      || space->purpose == FIL_TYPE_TEMPORARY
+		      || space->purpose == FIL_TYPE_IMPORT
+		      || my_atomic_loadlint(&space->redo_skipped_count) > 0
+		      || srv_is_tablespace_truncated(space->id));
 	}
 
 	ut_ad(space);
 	ut_ad(space->id == space_id);
-	x_lock_space(space, file, line);
+	x_lock(&space->latch, file, line);
 	ut_ad(space->purpose == FIL_TYPE_TEMPORARY
 	      || space->purpose == FIL_TYPE_IMPORT
 	      || space->purpose == FIL_TYPE_TABLESPACE);
 	return(space);
 }
 
-/** Exclusively aqcuire a tablespace latch.
-@param space  tablespace
-@param file   source code file name of the caller
-@param line   source code line number */
-void mtr_t::x_lock_space(fil_space_t *space, const char *file, unsigned line)
-{
-  rw_lock_x_lock_inline(&space->latch, 0, file, line);
-  memo_push(space, MTR_MEMO_SPACE_X_LOCK);
-}
-
-/** Look up the system tablespace. */
-void
-mtr_t::lookup_sys_space()
-{
-	ut_ad(!m_sys_space);
-	m_sys_space = fil_space_get(TRX_SYS_SPACE);
-	ut_ad(m_sys_space);
-}
-
-/** Look up the user tablespace.
-@param[in]	space_id	tablespace ID */
-void
-mtr_t::lookup_user_space(ulint space_id)
-{
-	ut_ad(space_id != TRX_SYS_SPACE);
-	ut_ad(m_user_space_id == space_id);
-	ut_ad(!m_user_space);
-	m_user_space = fil_space_get(space_id);
-	ut_ad(m_user_space);
-}
-
-/** Set the tablespace associated with the mini-transaction
-(needed for generating a MLOG_FILE_NAME record)
-@param[in]	space	user or system tablespace */
-void
-mtr_t::set_named_space(fil_space_t* space)
-{
-	ut_ad(m_user_space_id == TRX_SYS_SPACE);
-	ut_d(m_user_space_id = space->id);
-	if (space->id == TRX_SYS_SPACE) {
-		ut_ad(!m_sys_space || m_sys_space == space);
-		m_sys_space = space;
-	} else {
-		m_user_space = space;
-	}
-}
-
 /** Release an object in the memo stack.
 @return true if released */
 bool
 mtr_t::memo_release(const void* object, ulint type)
 {
+	ut_ad(m_impl.m_magic_n == MTR_MAGIC_N);
 	ut_ad(is_active());
 
 	/* We cannot release a page that has been written to in the
 	middle of a mini-transaction. */
-	ut_ad(!m_modifications || type != MTR_MEMO_PAGE_X_FIX);
+	ut_ad(!m_impl.m_modifications || type != MTR_MEMO_PAGE_X_FIX);
 
 	Iterate<Find> iteration(Find(object, type));
 
-	if (!m_memo.for_each_block_in_reverse(iteration)) {
+	if (!m_impl.m_memo.for_each_block_in_reverse(iteration)) {
 		memo_slot_release(iteration.functor.m_slot);
 		return(true);
 	}
@@ -743,15 +726,16 @@ mtr_t::memo_release(const void* object, ulint type)
 void
 mtr_t::release_page(const void* ptr, mtr_memo_type_t type)
 {
+	ut_ad(m_impl.m_magic_n == MTR_MAGIC_N);
 	ut_ad(is_active());
 
 	/* We cannot release a page that has been written to in the
 	middle of a mini-transaction. */
-	ut_ad(!m_modifications || type != MTR_MEMO_PAGE_X_FIX);
+	ut_ad(!m_impl.m_modifications || type != MTR_MEMO_PAGE_X_FIX);
 
 	Iterate<FindPage> iteration(FindPage(ptr, type));
 
-	if (!m_memo.for_each_block_in_reverse(iteration)) {
+	if (!m_impl.m_memo.for_each_block_in_reverse(iteration)) {
 		memo_slot_release(iteration.functor.get_slot());
 		return;
 	}
@@ -762,30 +746,37 @@ mtr_t::release_page(const void* ptr, mtr_memo_type_t type)
 
 /** Prepare to write the mini-transaction log to the redo log buffer.
 @return number of bytes to write in finish_write() */
-inline ulint mtr_t::prepare_write()
+ulint
+mtr_t::Command::prepare_write()
 {
 	ut_ad(!recv_no_log_write);
 
-	if (UNIV_UNLIKELY(m_log_mode != MTR_LOG_ALL)) {
-		ut_ad(m_log_mode == MTR_LOG_NO_REDO);
-		ut_ad(m_log.size() == 0);
+	switch (m_impl->m_log_mode) {
+	case MTR_LOG_SHORT_INSERTS:
+		ut_ad(0);
+		/* fall through */
+	case MTR_LOG_NO_REDO:
+	case MTR_LOG_NONE:
+		ut_ad(m_impl->m_log.size() == 0);
 		log_mutex_enter();
-		m_commit_lsn = log_sys->lsn;
-		return 0;
+		m_end_lsn = m_start_lsn = log_sys.lsn;
+		return(0);
+	case MTR_LOG_ALL:
+		break;
 	}
 
-	ulint	len	= m_log.size();
-	ulint	n_recs	= m_n_log_recs;
+	ulint	len	= m_impl->m_log.size();
+	ulint	n_recs	= m_impl->m_n_log_recs;
 	ut_ad(len > 0);
 	ut_ad(n_recs > 0);
 
-	if (len > log_sys->buf_size / 2) {
-		log_buffer_extend((len + 1) * 2);
+	if (len > srv_log_buffer_size / 2) {
+		log_buffer_extend(ulong((len + 1) * 2));
 	}
 
-	ut_ad(m_n_log_recs == n_recs);
+	ut_ad(m_impl->m_n_log_recs == n_recs);
 
-	fil_space_t*	space = m_user_space;
+	fil_space_t*	space = m_impl->m_user_space;
 
 	if (space != NULL && is_predefined_tablespace(space->id)) {
 		/* Omit MLOG_FILE_NAME for predefined tablespaces. */
@@ -794,32 +785,35 @@ inline ulint mtr_t::prepare_write()
 
 	log_mutex_enter();
 
-	if (fil_names_write_if_was_clean(space, this)) {
+	if (fil_names_write_if_was_clean(space, m_impl->m_mtr)) {
 		/* This mini-transaction was the first one to modify
 		this tablespace since the latest checkpoint, so
 		some MLOG_FILE_NAME records were appended to m_log. */
-		ut_ad(m_n_log_recs > n_recs);
-		mlog_catenate_ulint(&m_log, MLOG_MULTI_REC_END, MLOG_1BYTE);
-		len = m_log.size();
+		ut_ad(m_impl->m_n_log_recs > n_recs);
+		mlog_catenate_ulint(
+			&m_impl->m_log, MLOG_MULTI_REC_END, MLOG_1BYTE);
+		len = m_impl->m_log.size();
 	} else {
 		/* This was not the first time of dirtying a
 		tablespace since the latest checkpoint. */
 
-		ut_ad(n_recs == m_n_log_recs);
+		ut_ad(n_recs == m_impl->m_n_log_recs);
 
 		if (n_recs <= 1) {
 			ut_ad(n_recs == 1);
 
 			/* Flag the single log record as the
 			only record in this mini-transaction. */
-			*m_log.front()->begin() |= MLOG_SINGLE_REC_FLAG;
+			*m_impl->m_log.front()->begin()
+				|= MLOG_SINGLE_REC_FLAG;
 		} else {
 			/* Because this mini-transaction comprises
 			multiple log records, append MLOG_MULTI_REC_END
 			at the end. */
 
-			mlog_catenate_ulint(&m_log, MLOG_MULTI_REC_END,
-					    MLOG_1BYTE);
+			mlog_catenate_ulint(
+				&m_impl->m_log, MLOG_MULTI_REC_END,
+				MLOG_1BYTE);
 			len++;
 		}
 	}
@@ -831,105 +825,98 @@ inline ulint mtr_t::prepare_write()
 }
 
 /** Append the redo log records to the redo log buffer
-@param[in] len	number of bytes to write
-@return start_lsn */
-inline lsn_t mtr_t::finish_write(ulint len)
+@param[in] len	number of bytes to write */
+void
+mtr_t::Command::finish_write(
+	ulint	len)
 {
-	ut_ad(m_log_mode == MTR_LOG_ALL);
+	ut_ad(m_impl->m_log_mode == MTR_LOG_ALL);
 	ut_ad(log_mutex_own());
-	ut_ad(m_log.size() == len);
+	ut_ad(m_impl->m_log.size() == len);
 	ut_ad(len > 0);
 
-	lsn_t start_lsn;
-
-	if (m_log.is_small()) {
-		const mtr_buf_t::block_t* front = m_log.front();
+	if (m_impl->m_log.is_small()) {
+		const mtr_buf_t::block_t*	front = m_impl->m_log.front();
 		ut_ad(len <= front->used());
 
-		m_commit_lsn = log_reserve_and_write_fast(front->begin(), len,
-							  &start_lsn);
+		m_end_lsn = log_reserve_and_write_fast(
+			front->begin(), len, &m_start_lsn);
 
-		if (m_commit_lsn) {
-			return start_lsn;
+		if (m_end_lsn > 0) {
+			return;
 		}
 	}
 
 	/* Open the database log for log_write_low */
-	start_lsn = log_reserve_and_open(len);
+	m_start_lsn = log_reserve_and_open(len);
 
 	mtr_write_log_t	write_log;
-	m_log.for_each_block(write_log);
+	m_impl->m_log.for_each_block(write_log);
 
-	m_commit_lsn = log_close();
-	return start_lsn;
+	m_end_lsn = log_close();
 }
 
-/** Release the free extents that was reserved using
-fsp_reserve_free_extents().  This is equivalent to calling
-fil_space_release_free_extents().  This is intended for use
-with index pages.
-@param[in]	n_reserved	number of reserved extents */
+/** Release the latches and blocks acquired by this mini-transaction */
 void
-mtr_t::release_free_extents(ulint n_reserved)
+mtr_t::Command::release_all()
 {
-  fil_space_t *space= m_user_space;
+	m_impl->m_memo.for_each_block_in_reverse(CIterate<ReleaseAll>());
 
-  ut_ad(!m_undo_space);
-
-  if (space)
-    ut_ad(m_user_space->id == m_user_space_id);
-  else
-  {
-    ut_ad(m_sys_space->id == TRX_SYS_SPACE);
-    space= m_sys_space;
-  }
-
-  ut_ad(memo_contains(get_memo(), space, MTR_MEMO_SPACE_X_LOCK));
-  space->release_free_extents(n_reserved);
+	/* Note that we have released the latches. */
+	m_locks_released = 1;
 }
 
-/** Find out whether a block was not X-latched by the mini-transaction */
-struct FindBlockX
+/** Release the latches acquired by this mini-transaction */
+void
+mtr_t::Command::release_latches()
 {
-  const buf_block_t &block;
+	m_impl->m_memo.for_each_block_in_reverse(CIterate<ReleaseLatches>());
 
-  FindBlockX(const buf_block_t &block): block(block) {}
+	/* Note that we have released the latches. */
+	m_locks_released = 1;
+}
 
-  /** @return whether the block was not found x-latched */
-  bool operator()(const mtr_memo_slot_t *slot) const
-  {
-    return slot->object != &block || slot->type != MTR_MEMO_PAGE_X_FIX;
-  }
-};
-
-#ifdef UNIV_DEBUG
-/** Assert that the block is not present in the mini-transaction */
-struct FindNoBlock
+/** Release the blocks used in this mini-transaction */
+void
+mtr_t::Command::release_blocks()
 {
-  const buf_block_t &block;
+	m_impl->m_memo.for_each_block_in_reverse(
+		CIterate<const ReleaseBlocks>(
+			ReleaseBlocks(m_start_lsn, m_end_lsn,
+				      m_impl->m_flush_observer)));
+}
 
-  FindNoBlock(const buf_block_t &block): block(block) {}
-
-  /** @return whether the block was not found */
-  bool operator()(const mtr_memo_slot_t *slot) const
-  {
-    return slot->object != &block;
-  }
-};
-#endif /* UNIV_DEBUG */
-
-bool mtr_t::have_x_latch(const buf_block_t &block) const
+/** Write the redo log record, add dirty pages to the flush list and release
+the resources. */
+void
+mtr_t::Command::execute()
 {
-  if (m_memo.for_each_block(CIterate<FindBlockX>(FindBlockX(block))))
-  {
-    ut_ad(m_memo.for_each_block(CIterate<FindNoBlock>(FindNoBlock(block))));
-    ut_ad(!memo_contains_flagged(&block,
-                                 MTR_MEMO_PAGE_S_FIX | MTR_MEMO_PAGE_SX_FIX |
-                                 MTR_MEMO_BUF_FIX | MTR_MEMO_MODIFY));
-    return false;
-  }
-  ut_ad(rw_lock_own(&block.lock, RW_LOCK_X));
-  return true;
+	ut_ad(m_impl->m_log_mode != MTR_LOG_NONE);
+
+	if (const ulint len = prepare_write()) {
+		finish_write(len);
+	}
+
+	if (m_impl->m_made_dirty) {
+		log_flush_order_mutex_enter();
+	}
+
+	/* It is now safe to release the log mutex because the
+	flush_order mutex will ensure that we are the first one
+	to insert into the flush list. */
+	log_mutex_exit();
+
+	m_impl->m_mtr->m_commit_lsn = m_end_lsn;
+
+	release_blocks();
+
+	if (m_impl->m_made_dirty) {
+		log_flush_order_mutex_exit();
+	}
+
+	release_latches();
+
+	release_resources();
 }
 
 #ifdef UNIV_DEBUG
@@ -946,17 +933,15 @@ mtr_t::memo_contains(
 		return(false);
 	}
 
-	const rw_lock_t *lock = static_cast<const rw_lock_t*>(object);
-
 	switch (type) {
 	case MTR_MEMO_X_LOCK:
-		ut_ad(rw_lock_own(lock, RW_LOCK_X));
+		ut_ad(rw_lock_own((rw_lock_t*) object, RW_LOCK_X));
 		break;
 	case MTR_MEMO_SX_LOCK:
-		ut_ad(rw_lock_own(lock, RW_LOCK_SX));
+		ut_ad(rw_lock_own((rw_lock_t*) object, RW_LOCK_SX));
 		break;
 	case MTR_MEMO_S_LOCK:
-		ut_ad(rw_lock_own(lock, RW_LOCK_S));
+		ut_ad(rw_lock_own((rw_lock_t*) object, RW_LOCK_S));
 		break;
 	}
 
@@ -1030,9 +1015,10 @@ struct FlaggedCheck {
 bool
 mtr_t::memo_contains_flagged(const void* ptr, ulint flags) const
 {
-	ut_ad(is_active());
+	ut_ad(m_impl.m_magic_n == MTR_MAGIC_N);
+	ut_ad(is_committing() || is_active());
 
-	return !m_memo.for_each_block_in_reverse(
+	return !m_impl.m_memo.for_each_block_in_reverse(
 		CIterate<FlaggedCheck>(FlaggedCheck(ptr, flags)));
 }
 
@@ -1048,7 +1034,7 @@ mtr_t::memo_contains_page_flagged(
 	ulint		flags) const
 {
 	Iterate<FindPage> iteration(FindPage(ptr, flags));
-	return m_memo.for_each_block_in_reverse(iteration)
+	return m_impl.m_memo.for_each_block_in_reverse(iteration)
 		? NULL : iteration.functor.get_block();
 }
 
@@ -1071,7 +1057,7 @@ void
 mtr_t::print() const
 {
 	ib::info() << "Mini-transaction handle: memo size "
-		<< m_memo.size() << " bytes log size "
+		<< m_impl.m_memo.size() << " bytes log size "
 		<< get_log()->size() << " bytes";
 }
 
