@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2007, 2015, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2021, MariaDB Corporation.
+Copyright (c) 2017, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -44,8 +44,8 @@ Created July 17, 2007 Vasil Dimov
 #include "sync0rw.h"
 #include "sync0sync.h"
 #include "trx0sys.h"
-
-#include <sql_class.h>
+#include "que0que.h"
+#include "trx0purge.h"
 
 /** Initial number of rows in the table cache */
 #define TABLE_CACHE_INITIAL_ROWSNUM	1024
@@ -161,10 +161,10 @@ struct trx_i_s_cache_t {
 	ha_storage_t*	storage;	/*!< storage for external volatile
 					data that may become unavailable
 					when we release
-					lock_sys->mutex or trx_sys->mutex */
+					lock_sys.mutex or trx_sys.mutex */
 	ulint		mem_allocd;	/*!< the amount of memory
 					allocated with mem_alloc*() */
-	ibool		is_truncated;	/*!< this is TRUE if the memory
+	bool		is_truncated;	/*!< this is true if the memory
 					limit was hit and thus the data
 					in the cache is truncated */
 };
@@ -450,6 +450,7 @@ fill_trx_row(
 						which to copy volatile
 						strings */
 {
+	size_t		stmt_len;
 	const char*	s;
 
 	ut_ad(lock_mutex_own());
@@ -484,14 +485,16 @@ fill_trx_row(
 	row->trx_mysql_thread_id = thd_get_thread_id(trx->mysql_thd);
 
 	char	query[TRX_I_S_TRX_QUERY_MAX_LEN + 1];
-	if (size_t stmt_len = thd_query_safe(trx->mysql_thd, query,
-					     sizeof query)) {
+	stmt_len = innobase_get_stmt_safe(trx->mysql_thd, query, sizeof(query));
+
+	if (stmt_len > 0) {
+
 		row->trx_query = static_cast<const char*>(
 			ha_storage_put_memlim(
 				cache->storage, query, stmt_len + 1,
 				MAX_ALLOWED_FOR_STORAGE(cache)));
 
-		row->trx_query_cs = thd_charset(trx->mysql_thd);
+		row->trx_query_cs = innobase_get_charset(trx->mysql_thd);
 
 		if (row->trx_query == NULL) {
 
@@ -523,9 +526,9 @@ thd_done:
 
 	row->trx_tables_locked = lock_number_of_tables_locked(&trx->lock);
 
-	/* These are protected by both trx->mutex or lock_sys->mutex,
-	or just lock_sys->mutex. For reading, it suffices to hold
-	lock_sys->mutex. */
+	/* These are protected by both trx->mutex or lock_sys.mutex,
+	or just lock_sys.mutex. For reading, it suffices to hold
+	lock_sys.mutex. */
 
 	row->trx_lock_structs = UT_LIST_GET_LEN(trx->lock.trx_locks);
 
@@ -577,7 +580,7 @@ thd_done:
 
 	row->trx_is_read_only = trx->read_only;
 
-	row->trx_is_autocommit_non_locking = trx->is_autocommit_non_locking();
+	row->trx_is_autocommit_non_locking = trx_is_autocommit_non_locking(trx);
 
 	return(TRUE);
 }
@@ -596,7 +599,7 @@ put_nth_field(
 	ulint			n,	/*!< in: number of field */
 	const dict_index_t*	index,	/*!< in: index */
 	const rec_t*		rec,	/*!< in: record */
-	const rec_offs*		offsets)/*!< in: record offsets, returned
+	const ulint*		offsets)/*!< in: record offsets, returned
 					by rec_get_offsets() */
 {
 	const byte*	data;
@@ -677,8 +680,8 @@ fill_lock_data(
 	const dict_index_t*	index;
 	ulint			n_fields;
 	mem_heap_t*		heap;
-	rec_offs		offsets_onstack[REC_OFFS_NORMAL_SIZE];
-	rec_offs*		offsets;
+	ulint			offsets_onstack[REC_OFFS_NORMAL_SIZE];
+	ulint*			offsets;
 	char			buf[TRX_I_S_LOCK_DATA_MAX_LEN];
 	ulint			buf_used;
 	ulint			i;
@@ -1219,116 +1222,64 @@ trx_i_s_cache_clear(
 	ha_storage_empty(&cache->storage);
 }
 
-/*******************************************************************//**
-Fetches the data needed to fill the 3 INFORMATION SCHEMA tables into the
-table cache buffer. Cache must be locked for write. */
-static
-void
-fetch_data_into_cache_low(
-/*======================*/
-	trx_i_s_cache_t*	cache,		/*!< in/out: cache */
-	bool			read_write,	/*!< in: only read-write
-						transactions */
-	trx_ut_list_t*		trx_list)	/*!< in: trx list */
+
+/**
+  Add transactions to innodb_trx's cache.
+
+  We also add all locks that are relevant to each transaction into
+  innodb_locks' and innodb_lock_waits' caches.
+*/
+
+static void fetch_data_into_cache_low(trx_i_s_cache_t *cache, const trx_t *trx)
 {
-	const trx_t*		trx;
-	bool			rw_trx_list = trx_list == &trx_sys->rw_trx_list;
+  i_s_locks_row_t *requested_lock_row;
 
-	ut_ad(rw_trx_list || trx_list == &trx_sys->mysql_trx_list);
+  assert_trx_nonlocking_or_in_list(trx);
 
-	/* Iterate over the transaction list and add each one
-	to innodb_trx's cache. We also add all locks that are relevant
-	to each transaction into innodb_locks' and innodb_lock_waits'
-	caches. */
+  if (add_trx_relevant_locks_to_cache(cache, trx, &requested_lock_row))
+  {
+    if (i_s_trx_row_t *trx_row= reinterpret_cast<i_s_trx_row_t*>(
+        table_cache_create_empty_row(&cache->innodb_trx, cache)))
+    {
+      if (fill_trx_row(trx_row, trx, requested_lock_row, cache))
+        return;
+      --cache->innodb_trx.rows_used;
+    }
+  }
 
-	for (trx = UT_LIST_GET_FIRST(*trx_list);
-	     trx != NULL;
-	     trx =
-	     (rw_trx_list
-	      ? UT_LIST_GET_NEXT(trx_list, trx)
-	      : UT_LIST_GET_NEXT(mysql_trx_list, trx))) {
-
-		i_s_trx_row_t*		trx_row;
-		i_s_locks_row_t*	requested_lock_row;
-
-		/* Note: Read only transactions that modify temporary
-		tables an have a transaction ID */
-		if (!trx_is_started(trx)
-		    || (!rw_trx_list && trx->id != 0 && !trx->read_only)) {
-
-			continue;
-		}
-
-
-#ifdef UNIV_DEBUG
-		if (trx->is_autocommit_non_locking()) {
-			ut_ad(trx->read_only);
-			ut_ad(!trx->is_recovered);
-			ut_ad(trx->mysql_thd);
-			ut_ad(trx->in_mysql_trx_list);
-			const trx_state_t state = trx->state;
-			ut_ad(state == TRX_STATE_NOT_STARTED
-			      || state == TRX_STATE_ACTIVE);
-		}
-		else {
-			ut_ad(trx->state != TRX_STATE_NOT_STARTED);
-		}
-#endif /* UNIV_DEBUG */
-
-		ut_ad(trx->in_rw_trx_list == rw_trx_list);
-
-		if (!add_trx_relevant_locks_to_cache(cache, trx,
-						     &requested_lock_row)) {
-
-			cache->is_truncated = TRUE;
-			return;
-		}
-
-		trx_row = reinterpret_cast<i_s_trx_row_t*>(
-			table_cache_create_empty_row(
-				&cache->innodb_trx, cache));
-
-		/* memory could not be allocated */
-		if (trx_row == NULL) {
-
-			cache->is_truncated = TRUE;
-			return;
-		}
-
-		if (!fill_trx_row(trx_row, trx, requested_lock_row, cache)) {
-
-			/* memory could not be allocated */
-			--cache->innodb_trx.rows_used;
-			cache->is_truncated = TRUE;
-			return;
-		}
-	}
+  /* memory could not be allocated */
+  cache->is_truncated= true;
 }
 
-/*******************************************************************//**
-Fetches the data needed to fill the 3 INFORMATION SCHEMA tables into the
-table cache buffer. Cache must be locked for write. */
-static
-void
-fetch_data_into_cache(
-/*==================*/
-	trx_i_s_cache_t*	cache)	/*!< in/out: cache */
+
+/**
+  Fetches the data needed to fill the 3 INFORMATION SCHEMA tables into the
+  table cache buffer. Cache must be locked for write.
+*/
+
+static void fetch_data_into_cache(trx_i_s_cache_t *cache)
 {
-	ut_ad(lock_mutex_own());
-	ut_ad(trx_sys_mutex_own());
+  ut_ad(lock_mutex_own());
+  trx_i_s_cache_clear(cache);
 
-	trx_i_s_cache_clear(cache);
-
-	/* Capture the state of the read-write transactions. This includes
-	internal transactions too. They are not on mysql_trx_list */
-	fetch_data_into_cache_low(cache, true, &trx_sys->rw_trx_list);
-
-	/* Capture the state of the read-only active transactions */
-	fetch_data_into_cache_low(cache, false, &trx_sys->mysql_trx_list);
-
-	cache->is_truncated = FALSE;
+  /* Capture the state of transactions */
+  mutex_enter(&trx_sys.mutex);
+  for (const trx_t *trx= UT_LIST_GET_FIRST(trx_sys.trx_list);
+       trx != NULL;
+       trx= UT_LIST_GET_NEXT(trx_list, trx))
+  {
+    if (trx_is_started(trx) && trx != purge_sys.query->trx)
+    {
+      fetch_data_into_cache_low(cache, trx);
+      if (cache->is_truncated)
+        break;
+     }
+  }
+  mutex_exit(&trx_sys.mutex);
+  cache->is_truncated= false;
 }
 
+
 /*******************************************************************//**
 Update the transactions cache if it has not been read for some time.
 Called from handler/i_s.cc.
@@ -1346,13 +1297,7 @@ trx_i_s_possibly_fetch_data_into_cache(
 	/* We need to read trx_sys and record/table lock queues */
 
 	lock_mutex_enter();
-
-	trx_sys_mutex_enter();
-
 	fetch_data_into_cache(cache);
-
-	trx_sys_mutex_exit();
-
 	lock_mutex_exit();
 
 	/* update cache last read time */
@@ -1365,7 +1310,7 @@ trx_i_s_possibly_fetch_data_into_cache(
 Returns TRUE if the data in the cache is truncated due to the memory
 limit posed by TRX_I_S_MEM_LIMIT.
 @return TRUE if truncated */
-ibool
+bool
 trx_i_s_cache_is_truncated(
 /*=======================*/
 	trx_i_s_cache_t*	cache)	/*!< in: cache */
@@ -1409,7 +1354,7 @@ trx_i_s_cache_init(
 
 	cache->mem_allocd = 0;
 
-	cache->is_truncated = FALSE;
+	cache->is_truncated = false;
 }
 
 /*******************************************************************//**
